# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
""" Copyright (c) 2000-2003 LOGILAB S.A. (Paris, FRANCE).
 http://www.logilab.fr/ -- mailto:contact@logilab.fr

 basic checker for Python code
"""

__revision__ = "$Id: classes.py,v 1.4 2003/06/03 14:37:17 syt Exp $"

import inspect
from logilab.common.astng import get_names, FuncDef
from logilab.pylint.interfaces import IASTNGChecker
from logilab.pylint.checkers import BaseChecker, ASTNGCheckerHandler


def get_ancestor_for_method(manager, class_node, method):
    """ get the first class which implements <method>
    return the astng representation of the class
    """
    for classe in class_node.__bases__:
        if classe.__dict__.has_key(method):
            return manager.astng_from_class(classe)
        res = get_ancestor_for_method(manager, classe, method)
        if res is not None:
            return res

def get_ancestor_for_attribute(manager, class_node, attr):
    """ get the first class which as an instance attribute <attr>
    return the astng representation of the class
    """
    for classe in class_node.__bases__:
        astng = manager.astng_from_class(classe)
        if astng.instance_attrs.has_key(attr):
            return astng
        res = get_ancestor_for_attribute(manager, classe, attr)
        if res is not None:
            return res


def get_method_astng(manager, name, class_node):
    """ look for a method astng in the given class node or its ancestors """
    if class_node.locals().has_key(name):
        method_astng = class_node.locals()[name]
    else:
        class_node = get_ancestor_for_method(manager, class_node, name)
        if class_node is not None:
            # get astng for the searched method
            try:
                method_astng = class_node.locals()[name]
            except KeyError:
                # we have found the method but it's not in the local dictionnary
                # this may happen with astng build from living objects only
                return
    if not isinstance(method_astng, FuncDef):
        # FIXME : unexpected... 
        return
    return method_astng
    
class ClassChecker(BaseChecker, ASTNGCheckerHandler):
    """checks for :                                                            
    * methods without self as first argument                                   
    * overriden methods signature                                              
    * access only to existant members via self                                 
    * attributes not defined in the __init__ method                            
    * supported interfaces implementation                                      
    * unreachable code                                                         
    """
    
    __implements__ = (IASTNGChecker)#, IStatable)

    name = 'classes'
    priority = -2
    options = (('self-warning',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help' : 'print a warning method without self as first \
argument'}
                ),
               ('init',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check that attributes are defined in the \
__init__ method'}               
                ),
               ('members',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check that accessed members are defined'}
                ),               
               ('interfaces',
                {'type' : 'yn', 'metavar' : '<y_or_n>', 'default' : 1,
                 'help': 'check declared interfaces are truly implemented'}
                ),               
               )

    def visit_classdef(self, class_node, close=0):
        if close:
            # checks attributes are defined in __init__
            if self.config.init:
                for attr, node in class_node.instance_attrs.items():
                    frame = node.get_frame()
                    if frame.__name__ not in ('__init__', '__new__', 'setUp'):
                        msg = 'Attribute %s defined outside __init__'
                        self.add_message('warning', msg % attr, node.line)
            # check access to existant members
            if self.config.members:
                manager = self.linter.manager
                for attr, nodes in self._accessed.items():
                    # is it an instance attribute ?
                    if class_node.instance_attrs.has_key(attr):
                        continue
                    # or a class attribute ?
                    if class_node.locals().has_key(attr):
                        continue
                    # or an inherited method / attribute ?
                    if get_ancestor_for_method(manager, class_node, attr):
                        continue
                    if get_ancestor_for_attribute(manager, class_node, attr):
                        continue
                    msg = 'Access to undefined member %s' % attr
                    for node in nodes:
                        self.add_message('error', msg, node.line)
                
            del self._accessed
        else:
            self._accessed = {}
            if self.config.interfaces:
                if class_node.object:
                    self.check_interfaces(class_node)

            
    def visit_funcdef(self, node, close=0):
        is_method = node.is_method()
        # check first argument is self if this is actually a method
        if is_method:
            if close:
                del self._first_attr
                return
            manager = self.linter.manager
            if not node.args:
                self.add_message('error', 'Method has no argument')
            elif self.config.self_warning and (node.__name__ != '__new__'
                                               and node.args[0] != 'self'):
                msg = 'Method doesn\'t have self as first argument'
                self.add_message('error', msg)
                self._first_attr = node.args[0]
            else:
                self._first_attr = node.args[0]
                
            if node.__name__ == '__init__':
                return
            # check signature if the method overrload an herited method
            overriden = get_ancestor_for_method(manager, node.parent,
                                                node.__name__)
            if overriden is not None:
                manager = self.linter.manager
                # get astng for the searched method
                try:
                    astng = overriden.locals()[node.__name__]
                    self.check_signature(node, astng, 'overriden')
                except KeyError:
                    # we have found the method but it's not in the local
                    # dictionnary.
                    # This may happen with astng build from living objects only
                    pass
            # check if the method overrload an attribute
            overriden = get_ancestor_for_attribute(manager, node.parent,
                                                   node.__name__)
            if overriden:
                msg = 'Method hide an inherited attribute from %s'
                self.add_message('error', msg % overriden.__name__)
    
            
    def visit_ifstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
    def visit_elifstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
    def visit_whilestmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
    def visit_exceptstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
    def visit_raisestmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
    def visit_printstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
    def visit_returnstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
    def visit_yieldstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)            
    def visit_expressionstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
            
    def visit_forstmt(self, node):
        for name in get_names(node.in_tokens):
            self.handle_name(name, node)
            
    def visit_assertstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
        if node.msg_tokens:
            for name in get_names(node.msg_tokens):
                self.handle_name(name, node)
                
    def visit_execstmt(self, node):
        for name in get_names(node.tokens):
            self.handle_name(name, node)
        if node.in_tokens is not None:
            for name in get_names(node.in_tokens):
                self.handle_name(name, node)
                
    def visit_assignmentstmt(self, node):
        """ get accessed members from the rhs """
        for name in get_names(node.rhs_tokens):
            self.handle_name(name, node)


    def handle_name(self, name, node):
        """ check if the name handle an access to a class member """
        if hasattr(self, '_attrs') and hasattr(self, '_first_attr'):
            parts = name.split('.')
            if parts[0] == self._first_attr and len(parts) > 1:
                self._accessed.setdefault(parts[1], []).append(node)

    def check_interfaces(self, class_node):
        """ check that the given class node really implements declared
        interfaces
        """
        try:
            implements = class_node.object.__implements__
        except AttributeError:
            return
        manager = self.linter.manager
        if not type(implements) in (type(()), type([])):
            implements = (implements,)
        for iface in implements:
            iface_node = manager.astng_from_class(iface)
            iface_name = iface.__name__
            for member in dir(iface):
                member = getattr(iface, member)
                if not inspect.ismethod(member):
                    continue
                name = member.__name__
                # get class method astng
                method_astng = get_method_astng(manager, name, class_node)
                if method_astng is None:
                    msg = 'missing method %s from interface %s' 
                    self.add_message('error', msg % (name, iface_name))
                    continue
                if not method_astng.parent is class_node:
                    # don't go further if it is an inherited method
                    continue
                # get iface method astng
                imethod_astng = get_method_astng(manager, name, iface_node)
                if imethod_astng is None:
                    msg = 'Unable to get member %s from interface %s' 
                    self.add_message('warning', msg % (name, iface_name))
                    continue
                # check signature
                self.check_signature(method_astng, imethod_astng,
                                     '%s interface' % iface_name)
        
    
    def check_signature(self, method1, method2, type):
        """ check that the signature of the two given methods match """
        if len(method1.args) != len(method2.args):
            msg = 'Arguments number differs from %s method'
            self.add_message('warning', msg % type, method1.line)
        elif len(method1.func_defaults) != len(method2.func_defaults):
            msg = 'Signature differs from %s method'
            self.add_message('warning', msg % type, method1.line)
            
        
def register(linter):
    """ required method to auto register this checker """
    linter.register_checker(ClassChecker(linter))
